import time
import traceback
from typing import List

from ._get_model import download_model

from helpers.styler import Styler
from helpers.sourcemanager import SourceManager
from helpers.utils import BatchOptions, get_version, print_exc, print_verbose, run_verbose

__version__ = get_version()


def _pause(sec):
    print_verbose('Pausing for 3 seconds...')
    time.sleep(sec)
    print_verbose('Waking up!')


def batch_download(source_strings: List[str], rootdir: str, batchOptions: BatchOptions):
    """Batch downloads model from CivitAI one by one."""

    source_manager = SourceManager()

    for id in source_manager.parse_src(source_strings):

        iter = 0
        while True:
            try:
                download_model(
                    id=id,
                    dst_root_path=rootdir,
                    batchOptions=batchOptions
                )
                _pause(batchOptions.pause_time)
                break
            except Exception as e:
                print('---------')
                run_verbose(traceback.print_exc)
                print_exc(e, '\n')
                print('---------')
                _pause(batchOptions.pause_time)
                if iter < batchOptions.retry_count:
                    print(Styler.stylize(
                        'Retrying to download the current model...', color='info'))
                    iter += 1
                else:
                    print(Styler.stylize(
                        f'Max retry of {batchOptions.retry_count} reached. Skipping the current model...', color='info'))
                    break



import os
import re
from typing import Dict, List, Union

import argparse

from helpers.styler import Styler
from helpers.exceptions import InputException, UnexpectedException
from helpers.argparse import PwdAction, ConfirmAction, ColoredArgParser
from helpers.utils import get_version, print_verbose

__all__ = ['get_args']


def add_shared_option(par):
    par.add_argument(
        '--verbose', action=argparse.BooleanOptionalAction, help='Prints out traceback and other useful information.')
    par.add_argument(
        '-v', '--version', action='version', version=f'civitdl v{get_version()}', help='Prints out the version of the program.'
    )


parser = ColoredArgParser(
    prog='civitconfig',
    description="civitconfig is a cli program used to set configurations for the main program, civitdl.",
    formatter_class=argparse.RawTextHelpFormatter
)


subparsers = parser.add_subparsers(
    dest='subcommand',
    required=True,
    help='Choose one of the following subcommands: default, sorter, alias.')

default_parser = subparsers.add_parser(
    'default', help='Set a default value for one of the options below.\nIf no options are provided, default will print the current default.')

default_parser.add_argument('-i', '--max-images', metavar='INT', type=int,
                            help='Set the default max number of images to download per model.')

default_parser.add_argument('-s', '--sorter', metavar='NAME', type=str,
                            help='Set the default sorter given name of sorter (filepath not accepted).')

default_parser.add_argument('-k', '--api-key', action=PwdAction, type=str, nargs=0,
                            help='Prompts the user for their api key to use for model downloads that require users to log in.')

default_parser.add_argument('--with-prompt', action=argparse.BooleanOptionalAction,
                            help='Toggles default behavior on whether to download prompt alongside images.')

default_parser.add_argument('--limit-rate', type=str,
                            help='Set the default limit for the download speed/rate of resources downloaded from CivitAI. Set it to 0 to disable limit.'
                            )

default_parser.add_argument('--retry-count', type=int,
                            help='Set the default max number of times to retry downloading a model if it fails.'
                            )

default_parser.add_argument('--pause-time', type=float,
                            help='Set the default number of seconds to pause between each model\'s download'
                            )
default_parser.add_argument('--cache-mode', type=str,
                            help='Set the default cache mode. Valid modes are 0 and 1 for now. 2 is not implemented yet.')
default_parser.add_argument('--model-overwrite', action=argparse.BooleanOptionalAction,
                            help='Set the default behavior on whether to overwrite or skip models that are already downloaded at path. model=overwrite to overwrite model. no-model-overwrite to skip model.')
add_shared_option(default_parser)


sorter_parser = subparsers.add_parser(
    'sorter', help='Sorter-related subcommand.\nCurrently supports listing, adding and deleting sorters.\nIf no options are provided, sorter will list all available sorters.\nExample usage: civitdl 123456 ./lora --sorter mysorter.',
    formatter_class=argparse.RawTextHelpFormatter
)
sorter_group = sorter_parser.add_mutually_exclusive_group()
sorter_group.add_argument('-a', '--add', metavar=('NAME', 'FILEPATH'), type=str, nargs=2,
                          help='Add/save a new sorter to civitdl program.\nExample: civitconfig sorter --add mysorter ./custom/tags.py.')
sorter_group.add_argument('-d', '--delete', metavar='NAME', type=str,
                          help='Delete sorter based on name of the sorter.\nExample: civitconfig sorter --delete mysorter.')
add_shared_option(sorter_parser)


alias_parser = subparsers.add_parser(
    'alias',
    help='Alias-related subcommand.\nCurrently supports listing, adding and deleting aliases.\nIf no options are provided, alias will list all available aliases.\nExample usage: civitdl 123456 @lora.',
    formatter_class=argparse.RawTextHelpFormatter
)
alias_group = alias_parser.add_mutually_exclusive_group()
alias_group.add_argument('-a', '--add', metavar=('NAME', 'FILEPATH'), type=str, nargs=2,
                         help='Add a new alias to the civitdl program.\nExample: civitconfig alias --add @lora ./ComfyUI/models/loras.')
alias_group.add_argument('-d', '--delete', metavar=('NAME'), type=str,
                         help='Delete alias based on name of the alias.\nExample: civitconfig alias --delete @lora.')
add_shared_option(alias_parser)


config_parser = subparsers.add_parser(
    'settings',
    help='Subcommand related to the operations of civitconfig.',
    formatter_class=argparse.RawTextHelpFormatter
)
config_group = config_parser.add_mutually_exclusive_group()
config_group.add_argument(
    '-r', '--reset', action=ConfirmAction, help='Delete config and reinstall the default configuration.\nWARNING: DO NOT RUN THIS UNLESS YOU ARE SURE YOU WANT TO DELETE.')
config_group.add_argument(
    '-d', '--download', metavar=('PATH'), type=str, help='Downloads config directory to specified path.')
add_shared_option(config_parser)


def get_args():
    parser_result = parser.parse_args()

    return vars(parser_result)



from dataclasses import InitVar, dataclass, field
from datetime import datetime
import json
import os
import shutil
import sys
import time
import math
import csv
import importlib.metadata
from typing import IO, Callable, Dict, Iterable, List, Literal, Union, Optional
import concurrent.futures
import requests
from tqdm import tqdm

from helpers.sorter.utils import import_sort_model
from helpers.sorter import basic, tags
from helpers.styler import Styler
from helpers.exceptions import CustomException, InputException, NotImplementedException, UnexpectedException
from helpers.validation import Validation

# Level 0

_verbose = False


def get_verbose():
    return _verbose


def set_verbose(verbose: bool):
    global _verbose
    _verbose = verbose
    return _verbose


def getDate():
    now = datetime.now()
    return now.strftime("%Y-%m-%d--%H:%M:%S-%f")


def get_version():
    return importlib.metadata.version('civitdl')


# Level 1 - Currently or in the future might depends on level 0


def run_verbose(fn, *args, **kwargs):
    if get_verbose():
        fn(*args, **kwargs)


def print_verbose(*args, **kwargs):
    if get_verbose():
        args = [Styler.stylize(str(arg), bg_color='info') for arg in args]
        print(*args, **kwargs)


def print_exc(exc: Exception, *args, **kwargs):
    if isinstance(exc, CustomException):
        print(exc, file=sys.stderr, *args, **kwargs)
    else:
        print(Styler.stylize(str(exc), color='exception'), *args,
              file=sys.stderr, **kwargs)


def safe_run(callback: Callable[..., any], *values: any) -> dict:
    try:
        return {"success": True, "data": callback(*values)}
    except Exception as e:
        return {"success": False, "error": e}


def concurrent_request(req_fn, urls, max_workers=16):
    res_list = None

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(req_fn, url) for url in urls]
        res_list = [future.result() for future in futures]

    return res_list


def get_progress_bar(total: float, desc: str):
    return tqdm(total=total, desc=desc,
                unit='iB', unit_scale=True)


def find_in_list(li, cond_fn: Callable[[any, int], bool], default=None):
    """Given a list and a condition function, where the first argument is the index of the item and the second argument is the value of the item, it returns the first value in the list when the condition is true"""
    return next((item for i, item in enumerate(li) if cond_fn(item, int)), default)


def createDirsIfNotExist(dirpaths):
    for dirpath in dirpaths:
        os.makedirs(dirpath, exist_ok=True)


def write_contents(file: IO, content_chunks: Iterable, limit_rate: Union[int, None] = None, update_pb: Union[Callable[[int], None], None] = None):
    last_chunk_time = time.perf_counter()
    for content in content_chunks:
        file.write(content)
        bytes_downloaded = len(content)

        download_time = time.perf_counter() - last_chunk_time
        speed = bytes_downloaded / download_time
        if limit_rate is not None and limit_rate is not 0 and speed > limit_rate:
            time_to_sleep = (bytes_downloaded / limit_rate) - download_time
            if time_to_sleep > 0:
                time.sleep(time_to_sleep)

        if update_pb:
            update_pb(bytes_downloaded)
        last_chunk_time = time.perf_counter()


def delete_file_if_exists(file_path):
    if os.path.exists(file_path):
        try:
            os.remove(file_path)
        except OSError as e:
            raise UnexpectedException(f"Error deleting file {file_path}: {e}")


def read_dict_from_csv(filepath: str):
    csv_dict = {}

    with open(filepath, 'r') as f:
        csv_iterator = csv.reader(f)
        next(csv_iterator)
        for row in csv_iterator:
            try:
                row[2]
                raise InputException(
                    f'CSV file is invalid. File contains rows with more than two columns.\nThe filepath for the invalid csv is "{filepath}"')
            except:
                csv_dict[row[0]] = row[1]

    return csv_dict

# Level 2 - Currently or in the future might depends on level 0 and 1


def write_to_file(filepath: str, content_chunks: Iterable, mode: str = None, limit_rate: Union[int, None] = 0, overwrite: bool = True, use_pb: bool = False, total: float = 0, desc: str = None):
    """Uses content_chunks to write to filepath bit by bit. If use_pb is enabled, it is recommended to set total kwarg to the length of the file to be written."""
    progress_bar = get_progress_bar(total, desc) if use_pb else None

    def update_progress_bar(bytes_downloaded):
        if progress_bar:
            progress_bar.update(bytes_downloaded)

    if not overwrite and os.path.exists(filepath):
        if (progress_bar):
            progress_bar.close()
        print(Styler.stylize(
            f'File already exists at "{filepath}"', color='info'))
    else:
        temp_dirpath = os.path.join(os.path.dirname(
            filepath), '.tmp')
        temp_filepath = os.path.join(temp_dirpath, os.path.basename(filepath))
        os.makedirs(temp_dirpath, exist_ok=True)
        try:
            with open(temp_filepath, mode if mode != None else 'w') as file:
                write_contents(file, content_chunks,
                               limit_rate, update_progress_bar)
            shutil.move(temp_filepath, filepath)
            shutil.rmtree(temp_dirpath)
        except Exception as e:
            shutil.rmtree(temp_dirpath)
            raise e
        if (progress_bar):
            progress_bar.close()


def write_to_files(dirpath: str, basenames: Iterable, content_chunks_list: Iterable[Iterable], mode: str = None, use_pb: bool = False, total: float = 0, desc: str = None):
    """Write content to multiple files in dirpath. If use_pb is enabled, it is recommended to set total kwarg to the number of files being written."""
    progress_bar = get_progress_bar(total, desc) if use_pb else None
    for basename, content_chunks in zip(basenames, content_chunks_list):
        filepath = os.path.join(dirpath, basename)
        with open(filepath, mode if mode != None else 'w') as file:
            write_contents(file, content_chunks)
            if (progress_bar):
                progress_bar.update(1)
    if (progress_bar):
        progress_bar.close()


def parse_bytes(size: Union[str, int, float], name: str):
    units = {"k": 10**3, "m": 10**6, "g": 10**9, "t": 10**12}

    res = safe_run(float, size)
    if res["success"] == True:
        number = round(res["data"])
        Validation.validate_integer(number, name)
        return number
    else:
        number, unit = (size[0:-1], size[-1:].lower())

        num_res = safe_run(float, number)
        if num_res["success"] == False:
            raise InputException(
                f'Invalid byte value for {name}: {size}')
        else:
            number = round(num_res["data"])
            Validation.validate_integer(number, name)

        if unit not in units:
            raise InputException(
                f'Invalid byte value for {name}: {unit}')
        else:
            return number * units[unit]


# Level 3 - Currently or in the future might depends on level 0, 1 and 2

class BatchOptions:
    sorter_name: str
    sorter: Callable[[Dict, Dict, str, str],
                     List[str]] = basic.sort_model
    max_images: int = 3
    api_key: Optional[str] = None

    with_prompt: bool = True
    limit_rate: int = 0
    retry_count: int = 3
    pause_time: int = 3

    cache_mode: Literal['0', '1', '2'] = '1'
    model_overwrite: bool = False

    verbose: Optional[bool] = None

    def __get_sorter(self, sorter: str):
        if not isinstance(sorter, property) and not isinstance(sorter, str):
            raise InputException(
                'Sorter provided is not a string in BatchOptions.')

        sorter_str = 'basic' if isinstance(sorter, property) else sorter
        if sorter_str == 'basic' or sorter_str == 'tags':
            self._sorter = tags.sort_model if sorter_str == 'tags' else basic.sort_model
        else:
            self._sorter = import_sort_model(sorter_str)

        print_verbose("Chosen Sorter Description: ", self._sorter.__doc__)
        return self._sorter

    def __init__(self, retry_count, pause_time, max_images, with_prompt, api_key, verbose, sorter, limit_rate, cache_mode, model_overwrite):
        self.session = requests.Session()

        if verbose is not None:
            Validation.validate_bool(verbose, 'verbose')
            self.verbose = verbose
            set_verbose(self.verbose)

        if sorter is not None:
            Validation.validate_string(sorter, 'sorter')
            self.sorter = self.__get_sorter(sorter)
            self.sorter_name = sorter

        if max_images is not None:
            Validation.validate_integer(max_images, 'max_images', min_value=0)
            self.max_images = max_images

        if api_key is not None and api_key is not '':
            Validation.validate_string(api_key, 'api_key')
            self.api_key = api_key

        if with_prompt is not None:
            Validation.validate_bool(with_prompt, 'with_prompt')
            self.with_prompt = with_prompt

        if limit_rate is not None:
            Validation.validate_types(
                limit_rate, [str, int, float], 'limit_rate')
            self.limit_rate = parse_bytes(limit_rate, "limit_rate")

        if retry_count is not None:
            Validation.validate_integer(
                retry_count, 'retry_count', min_value=0)
            self.retry_count = retry_count

        if pause_time is not None:
            Validation.validate_float(pause_time, 'pause_time', min_value=0)
            self.pause_time = pause_time

        if cache_mode is not None:
            Validation.validate_string(
                cache_mode, 'cache_mode', whitelist=['0', '1', '2'])
            if cache_mode is '2':
                raise NotImplementedException(
                    'cache mode of 2 has not been implemented yet')
            self.cache_mode = cache_mode

        if model_overwrite is not None:
            Validation.validate_bool(model_overwrite, 'model_overwrite')
            self.model_overwrite = model_overwrite

 # TODO: Where are we planning on using DefaultOptions and BatchOptions?... I think I should consider argparse and __main__ to be in the same context...


class DefaultOptions:
    sorter: Optional[str] = None
    max_images: Optional[int] = None
    api_key: Optional[str] = None

    with_prompt: Optional[bool] = None
    limit_rate: Optional[str] = None
    retry_count: Optional[int] = None
    pause_time: Optional[int] = None

    cache_mode: Optional[int] = None
    model_overwrite: Optional[bool] = None

    def __init__(self, sorter=None, max_images=None, api_key=None, with_prompt=None, limit_rate=None, retry_count=None, pause_time=None, cache_mode=None, model_overwrite=None):
        if sorter is not None:
            Validation.validate_string(
                sorter, 'sorter')

            self.sorter = sorter

        if max_images is not None:
            Validation.validate_integer(
                max_images, 'max_images', min_value=0
            )
            self.max_images = max_images

        if api_key is not None:
            Validation.validate_string(api_key, 'api_key')
            self.api_key = api_key

        if with_prompt is not None:
            Validation.validate_bool(with_prompt, 'with_prompt')
            self.with_prompt = with_prompt

        if limit_rate is not None:
            Validation.validate_string(limit_rate, 'limit_rate')
            parse_bytes(limit_rate, 'limit_rate')
            self.limit_rate = limit_rate

        if retry_count is not None:
            Validation.validate_integer(
                retry_count, 'retry_count', min_value=0
            )
            self.retry_count = retry_count

        if pause_time is not None:
            Validation.validate_float(
                pause_time, 'pause_time', min_value=0
            )
            self.pause_time = pause_time

        if cache_mode is not None:
            Validation.validate_string(
                cache_mode, 'cache_mode', whitelist=['0', '1', '2']
            )
            if cache_mode is '2':
                raise NotImplementedException(
                    'cache mode of 2 has not been implemented yet')
            self.cache_mode = cache_mode

        if model_overwrite is not None:
            Validation.validate_bool(model_overwrite, 'model_overwrite')
            self.model_overwrite = model_overwrite



from helpers.exceptions import InputException, UnexpectedException
from helpers.utils import BatchOptions, DefaultOptions
from .config import Config


class DefaultConfig(Config):

    def __init__(self, *args):
        super(DefaultConfig, self).__init__(*args)

    def __validate_valid_sorter(self, target_sorter_name):
        if len([True for [name, _, _] in self.getSortersList() if name == target_sorter_name]) == 0:
            raise InputException(
                f'Setting the default sorter {target_sorter_name} failed due to being undefined. Run "civitconfig sorter" to check for valid sorters.')

    def setDefault(self, default_options: DefaultOptions):
        config = self._getConfig()
        options = {key: value for key, value in
                   vars(default_options).items() if value is not None}
        valid_keys = self._get_valid_keys()

        for key, value in options.items():
            if key in valid_keys:
                if key == 'sorter':  # special case
                    self.__validate_valid_sorter(value)

                config['default'][key] = value
            else:
                raise UnexpectedException(
                    'Unknown key found while configuring default options!!!')
        self._saveConfig(config)



from typing import Callable, Dict, List
import importlib.util
import re
from dataclasses import dataclass

from helpers.validation import Validation
from helpers.constants import BLACKLISTED_DIR_CHARS


def import_sort_model(path: str) -> Callable[[Dict, Dict, str, str], List[str]]:
    spec = importlib.util.spec_from_file_location('sorter', path)
    sorter = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(sorter)
    return sorter.sort_model

# TODO: Add instruction for windows path


@dataclass
class SorterData:
    model_dir_path: str
    metadata_dir_path: str
    image_dir_path: str
    prompt_dir_path: str

    def __post_init__(self):

        self.model_dir_path = Validation.validate_dir_path(
            self.model_dir_path, 'sorter')

        self.metadata_dir_path = Validation.validate_dir_path(
            self.metadata_dir_path, 'sorter')

        self.image_dir_path = Validation.validate_dir_path(
            self.image_dir_path, 'sorter')

        self.prompt_dir_path = Validation.validate_dir_path(
            self.prompt_dir_path, 'sorter')


class DirName:
    @staticmethod
    def replace(string, char_dict):
        return ''.join([char_dict[char] if char in char_dict else char for char in string])

    @staticmethod
    def remove_extra_space(string: str):
        return ' '.join([part for part in string.split(' ') if part != ''])

    @staticmethod
    def remove_pun_before_specials(string: str, special_chars):
        return re.sub(fr"[\.?!,;:]+(?= *[{special_chars}])", "", string)

    # Replacements
    @classmethod
    def replace_with_rule_1(cls, string):
        dic = {}
        dic['<'] = '('
        dic['>'] = ')'
        dic[':'] = ','
        dic['"'] = '-'
        dic['/'] = '&'
        dic['\\'] = '&'
        dic['|'] = '&'
        dic['?'] = '.'
        dic['*'] = '&'
        return cls.remove_extra_space(cls.remove_pun_before_specials(cls.replace(string, dic), '\&\(\)\[\]\{\}\-'))

    @classmethod
    def replace_all_with_space(cls, string):
        dic = {char: ' ' for char in BLACKLISTED_DIR_CHARS}
        return cls.remove_extra_space(cls.replace(string, dic))

    @classmethod
    def replace_all_with_empty(cls, string):
        dic = {char: '' for char in BLACKLISTED_DIR_CHARS}
        return cls.remove_extra_space(cls.replace(string, dic))

    @classmethod
    def replace_all_with_dash(cls, string):
        dic = {char: '-' for char in BLACKLISTED_DIR_CHARS}
        return cls.remove_extra_space(cls.replace(string, dic))




import argparse
import getpass
from gettext import gettext
import sys

from helpers.styler import Styler


class PwdAction(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        mypass = getpass.getpass(prompt='Key: ')
        setattr(namespace, self.dest, mypass)


class ConfirmAction(argparse.BooleanOptionalAction):
    def __call__(self, parser, namespace, value, option_string=None):
        response = input(
            'Are you sure you want to move the current configuration to trash, and reset with default? (y/N): ').strip().lower()
        if response not in ('y', 'yes'):
            parser.error(f"Reset declined. Exiting.")

        super().__call__(parser, namespace, value, option_string)


class ColoredArgParser(argparse.ArgumentParser):

    def print_usage(self, file=None):
        if file is None:
            file = sys.stdout
        self._print_message(self.format_usage()[0].upper() +
                            self.format_usage()[1:],
                            file, 'warning')

    def print_help(self, file=None):
        if file is None:
            file = sys.stdout
        self._print_message(self.format_help()[0].upper() +
                            self.format_help()[1:],
                            file, 'info')

    def _print_message(self, message: str, file=None, color=None, styles=None):
        if message:
            if file is None:
                file = sys.stderr
            if color is None:
                file.write(message)
            else:
                file.write(
                    Styler.stylize(f'{message.strip()}\n', color=color, styles=styles))

    def exit(self, status=0, message=None):
        if message:
            self._print_message(message, sys.stderr,
                                'exception')
        sys.exit(status)

    def error(self, message):
        self.print_usage(sys.stderr)
        args = {'prog': self.prog, 'message': message}
        self.exit(2, gettext('%(prog)s: Error: %(message)s\n') % args)



from typing import Dict
import os

from civitdl.api.sorter import SorterData, DirName


def sort_model(model_dict: Dict, version_dict: Dict, filename: str, root_path: str):
    """Create nested directories with info about the specific model given root path and the model dict. If no tag is found in the model's json, the default dir path name used will be 'others'."""
    twodim_tags = [
        ['anime', 'others'],
        ['style', 'poses', 'clothings', 'character',
         'celebrity', 'concept', 'others']
    ]

    model_tags = model_dict['tags']
    path = os.path.join(root_path, version_dict['baseModel'].replace(' ', '_'))

    for tags in twodim_tags:
        matched_tags = list(set(tags) & set(model_tags))
        path = os.path.join(
            path, tags[-1] if len(matched_tags) == 0 else matched_tags[0])

    model_dir_name = DirName.replace_with_rule_1(model_dict['name'])
    model_dir_path = os.path.join(path, model_dir_name)
    extra_data_dir_path = os.path.join(
        model_dir_path, f'extra_data-vid_{version_dict["id"]}')

    return SorterData(
        model_dir_path=model_dir_path,
        metadata_dir_path=extra_data_dir_path,
        image_dir_path=extra_data_dir_path,
        prompt_dir_path=extra_data_dir_path
    )



from .styler import Styler
from .exceptions import CustomException, InputException, UnexpectedException
from .constants import BLACKLISTED_DIR_CHARS
import sys
import os
from typing import Iterable, Union


class Validation:
    @staticmethod
    def __validate_arg_name(arg_name):
        if not isinstance(arg_name, str):
            raise UnexpectedException(
                f'Name/title of validator is not a string (arg_name: {arg_name})'
            )

        if len(arg_name) == 0:
            raise UnexpectedException(
                f'Name/title of validator is not defined (arg_name: {arg_name}, length of arg_name: {len(arg_name)})')

    @staticmethod
    def __validate_range(min, max):
        if min is not None and not isinstance(min, (int, float)):
            raise UnexpectedException(
                f'Minimum number provided is not a number (min: {min}, type: {type(min)}).')

        if max is not None and not isinstance(max, (int, float)):
            raise UnexpectedException(
                f'Maximum number provided is not a number (max: {max}, type: {type(max)})')

        if max is not None and min is not None and max < min:
            raise UnexpectedException(
                f'Maximum is smaller than minimum allowed number in range (min: {min}, max: {max})')

    @classmethod
    def validate_string(cls, value, arg_name: str, min_len: Union[int, None] = None, max_len: Union[int, None] = None, whitelist: Iterable[str] = [], blacklist: Iterable[str] = []) -> str:
        # Validate keyword arguments

        cls.__validate_arg_name(arg_name)
        cls.__validate_range(min_len, max_len)
        if min_len is not None and min_len < 0:
            raise UnexpectedException(
                f'Minimum length provided with value "{value}" is below 0 (min_len: {min_len}).'
            )
        if max_len is not None and max_len < 0:
            raise UnexpectedException(
                f'Maximum length provided with value "{value}" is below 0 (max_len: {max_len}).'
            )

        if not isinstance(whitelist, Iterable):
            raise UnexpectedException(
                f'Whitelist provided with value "{value}" is not a list (whitelist: {whitelist}).')
        if not all(isinstance(item, str) for item in whitelist):
            raise UnexpectedException(
                f'Whitelist provided with value "{value}" contains non-string elements (whitelist: {whitelist}).'
            )

        if not isinstance(blacklist, Iterable):
            raise UnexpectedException(
                f'Blacklist provided with value "{value}" is not a list (blacklist: {blacklist}).')
        if not all(isinstance(item, str) for item in blacklist):
            raise UnexpectedException(
                f'Blacklist provided with value "{value}" contains non-string elements (blacklist: {blacklist}).'
            )
        if len(whitelist) > 0 and len(blacklist) > 0:
            raise UnexpectedException(
                f'Both whitelist and blacklist provided with value "{value}" is not empty. You can not have both whitelist and blacklist (whitelist: {whitelist}, blacklist: {blacklist}).'
            )

        # Validate value

        if not isinstance(value, str):
            raise InputException(
                f'Value provided for {arg_name} is not a string (value: {value}, type: {type(value)}).')

        if value == '':
            raise InputException(
                f'Value provided for {arg_name} is empty.')

        if min_len is not None and len(value) < min_len:
            raise InputException(
                f'Value provided for {arg_name} is below minimum length allowed (value: {value}, min_len: {min_len}).'
            )

        if max_len is not None and len(value) > max_len:
            raise InputException(
                f'Value provided for {arg_name} is above maximum length allowed (value: {value}, min_len: {min_len}).'
            )

        if whitelist != [] and value not in whitelist:
            raise InputException(
                f'Value provided for {arg_name} is not in whitelist (value: {value}, whitelist: {whitelist}).'
            )

        if blacklist != [] and value in blacklist:
            raise InputException(
                f'Value provided for {arg_name} is in blacklist (value: {value}, blacklist: {blacklist}).')

        return value

    @classmethod
    def validate_integer(cls, value, arg_name: str, min_value: Union[int, None] = None, max_value: Union[int, None] = None) -> int:
        # validate keyword arguments

        cls.__validate_arg_name(arg_name)
        cls.__validate_range(min_value, max_value)

        # Validate value

        if not isinstance(value, int):
            raise InputException(
                f'Value provided for {arg_name} is not an number (value: {value}, type: {type(value)}).')

        if min_value is not None and value < min_value:
            raise InputException(
                f'Value provided for {arg_name} is below the minimum number allowed (value: {value}, min_value: {min_value})')

        if max_value is not None and value > max_value:
            raise InputException(
                f'Value provided for {arg_name} is above the minimum number allowed (value: {value}, max_value: {max_value})')

        return value

    @classmethod
    def validate_float(cls, value, arg_name: str, min_value: Union[float, None] = None, max_value: Union[float, None] = None) -> float:
        # Validate keyword arguments

        cls.__validate_arg_name(arg_name)
        cls.__validate_range(min_value, max_value)

        # Validate value

        if not isinstance(value, float):
            raise InputException(
                f'Value provided for {arg_name} is not a number (value: {value}, type: {type(value)}).')

        if min_value is not None and value < min_value:
            raise InputException(
                f'Value provided for {arg_name} is below the minimum number allowed (value: {value}, min_value: {min_value})')

        if max_value is not None and value > max_value:
            raise InputException(
                f'Value provided for {arg_name} is above the minimum number allowed (value: {value}, max_value: {max_value})')

        return value

    @classmethod
    def validate_bool(cls, value, arg_name: str) -> bool:
        # Validate keyword arguments

        cls.__validate_arg_name(arg_name)

        # Validate value
        if not isinstance(value, bool):
            raise InputException(
                f'Value provided for {arg_name} is not a bool (value: {value}, type: {type(value)})')
        return value

    @classmethod
    def validate_types(cls, value, types, arg_name: str):
        # Validate keyword arguments

        cls.__validate_arg_name(arg_name)

        # Validate value
        if not (isinstance(value, tuple(types))):
            raise InputException(
                f'Value provided for {arg_name} is not one of the expected types (value: {value}, types: {types})')

        return value

    @classmethod
    def validate_dir_name(cls, value, arg_name):
        cls.__validate_arg_name(arg_name)

        # Validate value
        blacklist = ['<', '>', ':', '"', '/', '\\', '|', '?', '*']

        value = str(value)

        if len(value) == 0:
            raise InputException(
                f'Directory name provided by {arg_name} is empty.')

        if len(value.strip()) is not len(value):
            print(Styler.stylize(
                f'Directory name provided by {arg_name} contains leading and trailing space. Proceeding to trim the following "{value}"', color='WARNING'))
            value = value.trim()

        for el in value:
            if el in BLACKLISTED_DIR_CHARS:
                raise InputException(
                    f'Directory name provided by {arg_name} is invalid. It may not contain the illegal character, "{el}".',
                    f'The provided directory name is "{value}".',
                    f'The list of blacklisted characters are {BLACKLISTED_DIR_CHARS}.',
                    f'If this directory path is generated from a sorter, please report or change to a different sorter.'
                )

        return value

    @classmethod
    def validate_dir_path(cls, value, arg_name):
        cls.__validate_arg_name(arg_name)

        # Validate value
        value = str(value)

        if len(value) == 0:
            raise InputException(
                f'Directory path provided by {arg_name} is empty.')

        if len(value.strip()) is not len(value):
            print(Styler.stylize(
                f'Directory path provided by {arg_name} contains leading and trailing space. Proceeding to trim the following "{value}"', color='WARNING'))
            value = value.trim()

        modified_value = value.replace('/', '\\') if os.name == 'nt' else value
        dir_names = modified_value.split(os.path.sep)
        for dir_name in dir_names:
            try:
                cls.validate_dir_name(dir_name, arg_name)
            except InputException as e:
                raise InputException(
                    e, f'The invalid directory path is {value}')

        return modified_value



import os
import shutil
from helpers.exceptions import InputException, ResourcesException
from helpers.utils import getDate, print_exc
from helpers.sorter.utils import import_sort_model
from .config import Config


class SorterConfig(Config):

    def __init__(self, *args):
        super(SorterConfig, self).__init__(*args)

    def _copyPyFile(self, filepath) -> str:
        dst_filename = f'{getDate()}.py'
        dstpath = os.path.join(
            self._sorters_dir_path, dst_filename)
        shutil.copy2(filepath, dstpath)
        return dstpath

    def _uncopyPyFile(self, filepath):
        if os.path.exists(filepath):
            os.remove(filepath)
        else:
            raise ResourcesException(
                'Unable to uncopy python file (or delete it from program) because it does not exist')

    def _trashPyFile(self, filepath):
        filename = os.path.basename(filepath)
        trashpath = os.path.join(
            self._sorters_trash_dir_path, filename)
        if os.path.exists(filepath):
            shutil.move(filepath, trashpath)
            return trashpath
        else:
            raise ResourcesException(
                'Unable to trash python file because it does not exist.')

    def _untrashPyFile(self, trashpath):
        filename = os.path.basename(trashpath)
        filepath = os.path.join(
            self._sorters_dir_path, filename)
        if os.path.exists(trashpath):
            shutil.move(trashpath, filepath)
            return filepath
        else:
            raise ResourcesException(
                'Unable to untrash python file because it does not exist.')

    def trashPyFiles(self, pyfile_paths):
        trashfile_paths = []
        try:
            for pyfile_path in pyfile_paths:
                trashfile_paths.append(self._trashPyFile(pyfile_path))
            return trashfile_paths
        except Exception as e:
            for trashfile_paths in trashfile_paths:
                self._untrashPyFile(trashfile_paths)
            raise e

    def untrashPyFiles(self, trashfile_paths):
        pyfile_paths = []
        try:
            for trashfile_path in trashfile_paths:
                pyfile_paths.append(self._untrashPyFile(trashfile_path))
            return pyfile_paths
        except Exception as e:
            for pyfile_path in pyfile_paths:
                self._trashPyFile(pyfile_path)
            raise e

    def addSorter(self, name, filepath):
        sorters = self.getSortersList()

        if name == 'tags' or name == 'basic':
            raise InputException(f'Sorter with name "{name}" is reserved.')

        if '/' in name:
            raise InputException(f'Sorter name may not contain "/": {name}')

        if name in [sorter[0] for sorter in sorters]:
            raise InputException(f'Sorter with name "{name}" already exist.')

        desc = import_sort_model(
            filepath).__doc__ or "Description not provided"

        # First, we save python file (if exception, need to undo)
        filepath = self._copyPyFile(filepath)

        # Then, we edit config
        config = None
        try:
            config = self._getConfig()
            config['sorters'].append([name, desc, filepath])
        except:
            self._uncopyPyFile(filepath)
            raise e

        # Lastly, we save config
        try:
            self._saveConfig(config)
        except Exception as e:
            self._uncopyPyFile(filepath)
            raise e

    def deleteSorter(self, name):
        sorters = self.getSortersList()

        if name == 'tags' or name == 'basic':
            raise InputException(
                f'Sorter with name "{name}" can not be deleted.')

        if name not in [sorter[0] for sorter in sorters]:
            raise InputException(f'Sorter with name "{name}" does not exist.')

        sorter_py_path = [sorter[2]
                          for sorter in sorters if sorter[0] == name][0]

        # First, we edit config
        config = self._getConfig()
        config['sorters'] = [
            sorter for sorter in config['sorters'] if sorter[0] != name]

        # Then, we delete file (if exception, need to undo)
        resources_error_was_raised = False
        trashpath = None
        try:
            trashpath = self._trashPyFile(sorter_py_path)
        except Exception as e:
            if type(e).__name__ == 'ResourcesException':
                print_exc(e)
                resources_error_was_raised = True
            else:
                raise e

        # Lastly, we make changes to config
        try:
            self._saveConfig(config)
        except Exception as e:
            if not resources_error_was_raised:
                self._untrashPyFile(trashpath)
            raise e



from json import dumps
import shutil
import requests
import os
import re
from typing import Callable, Dict, List, Union
from math import ceil

from helpers.styler import Styler
from helpers.sourcemanager import Id
from helpers.utils import BatchOptions, delete_file_if_exists, write_to_file, write_to_files, print_verbose, concurrent_request
from helpers.sorter.utils import SorterData
from helpers.exceptions import InputException, ResourcesException, UnexpectedException, APIException
from helpers.validation import Validation
from helpers.hashmanager import HashManager


class Metadata:
    __batchOptions = None
    __id = None

    model_name = None
    model_id = None
    version_id = None
    version_hashes = None

    model_dict = None
    version_dict = None

    nsfw = False
    image_dicts = None
    image_download_urls = None
    model_download_url = None

    def __init__(self, id: Id, batchOptions: BatchOptions):
        self.__id = id
        self.__batchOptions = batchOptions
        self.__handler()

    def __handler(self):
        type = self.__id.type
        ids = self.__id.data

        if type == 'site' and len(ids) == 2:
            self.model_id = ids[0]
            self.version_id = ids[1]
            self.model_dict = self.__get_model_metadata()
            self.version_dict = self.__get_version_metadata()
        elif type == 'id' or (type == 'site' and len(ids) == 1):
            self.model_id = ids[0]
            self.model_dict = self.__get_model_metadata()
            if len(self.model_dict['modelVersions']) == 0:
                raise ResourcesException(
                    f'No model versions found from model id, {self.model_id}')
            self.version_dict = self.model_dict['modelVersions'][0]
            self.version_id = str(self.version_dict['id'])
        elif type == 'api':
            self.version_id = ids[0]
            self.version_dict = self.__get_version_metadata()
            self.model_id = str(self.version_dict['modelId'])
            self.model_dict = self.__get_model_metadata()
        else:
            raise InputException(
                f'Incorrect format sent ({ids}, {type}): "{self.__id.original}"')

        self.model_name = self.model_dict['name']

        self.nsfw = self.model_dict['nsfw']
        self.image_dicts = [
            image_dict for image_dict in self.version_dict['images']
            if self.nsfw or image_dict['nsfw'] == 'None'
        ][0:self.__batchOptions.max_images]
        self.image_download_urls = [image_dict['url']
                                    for image_dict in self.image_dicts]

        self.model_download_url = self.version_dict['downloadUrl']

        self.version_hashes = self.__get_version_hashes(
            self.version_dict, self.model_download_url)

    def __get_version_hashes(self, version_dict, download_url):
        files = version_dict['files']
        hashes = {}
        for file in files:
            if "downloadUrl" in file and file['downloadUrl'] == download_url:
                if isinstance(file['hashes'], dict):
                    hashes = file['hashes']
                else:
                    print(Styler.stylize(
                        'Hashes found in metadata is not a dictionary! There is an error with the API!', color='error'))
                break
        if hashes is {}:
            print(Styler.stylize('Hash not found in metadata.', color='warning'))
        elif "SHA256" not in hashes:
            print(Styler.stylize('SHA256 hash not found.', color='warning'))
        return hashes

    def __get_model_metadata(self):
        """Returns json object if request succeeds, else print error and returns None"""
        metadata_url = f'https://civitai.com/api/v1/models/{self.model_id}'
        return self.__get_metadata(metadata_url)

    def __get_version_metadata(self):
        metadata_url = f'https://civitai.com/api/v1/model-versions/{self.version_id}'
        return self.__get_metadata(metadata_url)

    def __get_metadata(self, url: str):
        print_verbose('Requesting model metadata.')
        print_verbose(f'Metadata API Request URL: {url}')
        meta_res = self.__batchOptions.session.get(url, stream=True)
        print_verbose('Finished requesting model metadata.')
        if meta_res.status_code != 200:
            raise APIException(
                meta_res.status_code, f'Downloading metadata from CivitAI for "{self.__id.original}" failed when trying to request metadata from "{url}"')

        try:
            return meta_res.json()
        except Exception as e:
            raise UnexpectedException(
                'Unable to parse metadata from CivitAI (incorrect format provided by Civitai).', 'CivitAI might be under maintainence.',
                f'\nOriginal Error:\n       {e}')


# TODO: what if a specific image have a hard time with getting a response?


def _download_images(dirpath: str, image_basenames: List[str], image_urls: List[str], batchOptions: BatchOptions):
    def make_req(url): return batchOptions.session.get(url, stream=True)

    # TODO: Change progress bar to be based on time length of request rather than when all the images are fetched and ready to be written.
    os.makedirs(dirpath, exist_ok=True)
    print_verbose('Now requesting images...')
    image_content_chunks_list = [res.iter_content(1024*1024) for res in concurrent_request(
        req_fn=make_req, urls=image_urls)]
    print_verbose('Finished requesting images...')

    if (len(image_content_chunks_list) == 0):
        print(Styler.stylize('No images to download...', color='warning'))
    else:
        write_to_files(dirpath, image_basenames, image_content_chunks_list, mode='wb',
                       use_pb=True, total=len(image_basenames), desc='Images')


def _download_prompts(dirpath: str, basenames: List[str], image_dicts: List[Dict]):
    if len(image_dicts) != 0:
        write_to_files(dirpath, basenames, [dumps(
            image_dict, indent=2, ensure_ascii=False) for image_dict in image_dicts])


def _download_metadata(dirpath: str, metadata: Metadata):

    model_dict_filename = f'model_dict-mid_{metadata.model_id}-vid_{metadata.version_id}.json'
    model_dict_path = os.path.join(
        dirpath, model_dict_filename)
    os.makedirs(dirpath, exist_ok=True)
    write_to_file(model_dict_path, [dumps(
        metadata.model_dict, indent=2, ensure_ascii=False)])


def _download_hashes(dirpath: str, filename_no_ext: str, hashes: Dict):
    hashes_dict_path = os.path.join(dirpath, filename_no_ext + '.csv')
    os.makedirs(dirpath, exist_ok=True)
    data = 'hash_name, hash_id\n'
    for key, value in hashes.items():
        data += f'{key}, {value}\n'
    write_to_file(hashes_dict_path, [data.rstrip()])


# TODO: Make it so api_key is only used when reason=download-auth is in res.url
def _get_filename_and_model_res(input_str: str, metadata: Metadata, batchOptions: BatchOptions):
    # Request model
    print_verbose('Preparing to send download model request...')
    print_verbose(f'Model Download API URL: {metadata.model_download_url}')
    headers = {
        'Authorization': f'Bearer {batchOptions.api_key}',
    } if batchOptions.api_key else {}
    res = batchOptions.session.get(metadata.model_download_url, stream=True, headers=headers) if batchOptions.api_key else requests.get(
        metadata.model_download_url, stream=True)
    print_verbose('Download model response received.')

    if res.status_code != 200:
        raise APIException(
            res.status_code, f'Downloading model from CivitAI failed for model id, {metadata.model_id}, and version id, {metadata.version_id}')

    # Find filename from content disposition
    content_disposition = res.headers.get('Content-Disposition')

    if 'reason=download-auth' in res.url:
        print_verbose('reason=download-auth status', res.status_code)
        raise InputException('Unable to download this model as it requires an API Key. Please head to "civitai.com", go to "Account Settings", then go to "API Keys" section, then add an api key to your account. After that, paste the key to the program.')

    if content_disposition == None:
        raise ResourcesException(
            f'Downloaded model from CivitAI has no content disposition header available.')
    filename = None

    try:
        filename = content_disposition.split(
            'filename=')[-1].strip('"').encode('latin-1').decode('utf-8')
    except UnicodeDecodeError:
        # Alternative solution for finding filename
        for file in metadata.version_dict['files']:
            file_version_id = re.search(
                r'(?<=models\/)\d+', file['downloadUrl'])
            if file_version_id != None and file_version_id.group(0) == metadata.version_id:
                filename = file['name']
                break

    if filename == None:
        raise UnexpectedException(
            f'Unable to retrieve filename for {input_str}')

    return (res, filename)


def download_model(id: Id, dst_root_path: str, batchOptions: BatchOptions):
    """
        Downloads the model's safetensors and json metadata files.
        create_dir_path is a callback function that takes in the following: metadata dict, specific model's data as dict, filename, and root path.
    """
    # if id == None or dst_root_path == None:
    #     raise InputException(
    #         '(download_model) Must provide an id, create_dir_path and dst_root_path')

    metadata = Metadata(id, batchOptions)

    print(Styler.stylize(
        f"""Now downloading \"{metadata.model_name}\"...
            - Model ID: {metadata.model_id}
            - Version ID: {metadata.version_id}\n""",
        color='main'))

    model_res, filename = _get_filename_and_model_res(
        id.original, metadata, batchOptions)

    # Create empty directory recursively #
    filename_no_ext, filename_ext = os.path.splitext(filename)
    sorter_data = batchOptions.sorter(
        metadata.model_dict, metadata.version_dict, filename_no_ext, dst_root_path)

    if not isinstance(sorter_data, SorterData):
        raise InputException(
            f'The following sorter, "{batchOptions.sorter_name}", returned the incorrect type.')

    # Write model to the directory #

    # Download metadata
    _download_metadata(sorter_data.metadata_dir_path, metadata)

    # Download images & prompts
    image_basenames = [os.path.basename(url)
                       for url in metadata.image_download_urls]
    prompts_basenames = [
        f'{os.path.splitext(basename)[0]}.json' for basename in image_basenames]

    _download_images(
        sorter_data.image_dir_path, image_basenames, metadata.image_download_urls, batchOptions)
    if batchOptions.with_prompt:
        _download_prompts(sorter_data.prompt_dir_path, prompts_basenames,
                          metadata.image_dicts)

    # Download model & hashes
    os.makedirs(sorter_data.model_dir_path, exist_ok=True)
    model_filename_no_ext = f'{filename_no_ext}-mid_{metadata.model_id}-vid_{metadata.version_id}'

    _download_hashes(sorter_data.model_dir_path,
                     model_filename_no_ext,
                     metadata.version_hashes)

    model_filename = model_filename_no_ext + filename_ext
    model_path = os.path.join(
        sorter_data.model_dir_path, model_filename)

    hash_manager = None
    cached_model_path = None
    if batchOptions.cache_mode is not '0':
        hash_manager = HashManager(metadata.version_id)
        cached_model_path = hash_manager.get_local_model_path()

    print(batchOptions.cache_mode)
    print(hash_manager)

    if hash_manager and cached_model_path and cached_model_path != model_path:
        print(Styler.stylize(f"""Model already existed at the following path:
            - Path: {cached_model_path}""", color='info'))
        print(Styler.stylize(f"Copying to new path...", color='info'))
        shutil.copy(cached_model_path, model_path)
    else:
        content_chunks = model_res.iter_content(
            ceil(batchOptions.limit_rate / 8)
            if batchOptions.limit_rate is not None and batchOptions.limit_rate is not 0
            else 1024*1024)
        write_to_file(model_path, content_chunks, mode='wb', limit_rate=batchOptions.limit_rate, overwrite=batchOptions.model_overwrite,
                      use_pb=True, total=float(model_res.headers.get('content-length', 0)), desc='Model')

    if hash_manager:
        hash_manager.set_local_model_cache(model_path, metadata.version_hashes)

    print(Styler.stylize(
        f"""\nDownload completed for \"{metadata.model_name}\" 
            - Model ID: {metadata.model_id}
            - Version ID: {metadata.version_id}
            - Path: {model_path}\n""", color='success'))

    print('---------------------------\n')



#!/usr/bin/env python3
import traceback
from operator import itemgetter

from .batch.batch_download import batch_download, BatchOptions
from .args.argparser import get_args

from helpers.exceptions import UnexpectedException
from helpers.sourcemanager import SourceManager
from helpers.utils import print_verbose, run_verbose, print_exc, set_verbose


def main():
    try:
        args = get_args()

        if args['verbose']:
            set_verbose(True)
        else:
            set_verbose(False)

        print_verbose('Arguments: ', str(args))

        batchOptions = BatchOptions(
            sorter=args['sorter'],
            max_images=args['max_images'],
            api_key=args['api_key'],

            with_prompt=args['with_prompt'],
            limit_rate=args['limit_rate'],
            retry_count=args['retry_count'],
            pause_time=args['pause_time'],

            cache_mode=args['cache_mode'],
            model_overwrite=args['model_overwrite'],

            verbose=args['verbose']
        )

        batch_download(
            source_strings=args['source_strings'],
            rootdir=args['rootdir'],
            batchOptions=batchOptions
        )

    except Exception as e:
        print('---------')
        run_verbose(traceback.print_exc)
        print_exc(e)
        print('---------')



import json
from .styler import Styler


class CustomException(Exception):
    def __init__(self, error_type, *messages):
        res = ""

        res += Styler.stylize(f'\n{self.__class__.__name__} ({error_type}):',
                              color='exception', styles=['bold'])

        for message in messages:
            res += Styler.stylize(f'\n         {message}', color='exception')

        super().__init__(res)


class InputException(CustomException):
    def __init__(self, *messages):
        super().__init__('Bad Inputs', *messages)


class ResourcesException(CustomException):
    def __init__(self, *messages):
        super().__init__('Resources Not Found', *messages)


class APIException(CustomException):
    def __init__(self, status_code, *messages):
        super().__init__(f'API Status Code {status_code}', *messages)


class NotImplementedException(CustomException):
    def __init__(self, *messages):
        super().__init__('Feature Not Implemented', *messages)


class UnexpectedException(CustomException):
    def __init__(self, *messages):
        super().__init__('Unexpected Error', *messages)



import os
from helpers.exceptions import InputException
from .config import Config


class AliasConfig(Config):

    def __init__(self, *args):
        super(AliasConfig, self).__init__(*args)

    def addAlias(self, alias_name: str, path: str):
        # if alias does not exist, we add
        aliases = self.getAliasesList()

        if '/' in alias_name:
            raise InputException(
                f'Alias name may not contain "/": {alias_name}')

        if alias_name in [aname for aname, _ in aliases]:
            raise InputException(
                f'Alias name exist already: ${alias_name}')

        config = self._getConfig()
        if path.split(os.path.sep)[0] in [aname for aname, _ in aliases]:
            config['aliases'].append([alias_name, path])
        else:
            config['aliases'].append([alias_name, os.path.abspath(path)])

        self._saveConfig(config)

    def deleteAlias(self, alias_name: str):
        aliases = self.getAliasesList()

        if alias_name not in [aname for aname, _ in aliases]:
            raise InputException(
                f'Alias with name {alias_name} does not exist.')

        config = self._getConfig()
        config['aliases'] = [
            alias for alias in config['aliases'] if alias[0] != alias_name
        ]

        self._saveConfig(config)




import os
import json
from datetime import datetime
import shutil
from typing import Dict, List, Union

from helpers.sorter import basic, tags
from helpers.exceptions import UnexpectedException
from helpers.styler import Styler
from helpers.utils import getDate
from helpers.validation import Validation

DEFAULT_CONFIG = {
    "version": "1",
    "default": {
        "sorter": "basic",
        "max_images": 3,
        "api_key": "",

        "with_prompt": True,
        "limit_rate": '0',
        "retry_count": 3,
        "pause_time": 3.0,

        "cache_mode": '1',
        "model_overwrite": False
    },
    "sorters": [["basic", basic.sort_model.__doc__, 'basic'], ["tags", tags.sort_model.__doc__, 'tags']],
    "aliases": [["@example", "~/.models"]]
}


CURRENT_VERSION = "1"

_config = None  # Do not do async operations with this


class Config:

    _config_path: str
    _config_dir_path: str
    _config_trash_dir_path: str
    _sorters_dir_path: str
    _sorters_trash_dir_path: str

    def __init__(self, config_path, config_dir_path, config_trash_dir_path, sorters_dir_path, sorters_trash_dir_path):
        self._config_path = config_path
        self._config_dir_path = config_dir_path
        self._config_trash_dir_path = config_trash_dir_path
        self._sorters_dir_path = sorters_dir_path
        self._sorters_trash_dir_path = sorters_trash_dir_path

    # Private Methods #

    ## Configs ##

    def _configExists(self):
        return os.path.exists(self._config_path)

    def _getConfig(self):
        global _config
        if not _config:
            with open(self._config_path) as file:
                _config = json.load(file)
            if _config == None:
                raise UnexpectedException('JSON config file was not read...')

        return _config

    def _saveConfig(self, dic: Dict):
        global _config
        with open(self._config_path, 'w') as file:
            json.dump(dic, file, indent=2)
            _config = dic

    def _get_valid_keys(self):
        return DEFAULT_CONFIG['default'].keys()

    # Public Methods #

    def getDefault(self, key=None):
        config = self._getConfig()
        if key == None:
            return config['default']
        else:
            Validation.validate_string(key, 'getDefault', min_len=1)
            return config['default'][key]

    def getSortersList(self) -> List:
        return self._getConfig()['sorters']

    def getAliasesList(self) -> List:
        return self._getConfig()['aliases']

    def print_defaults(self):
        blacklisted_keys = ['api_key']
        print(Styler.stylize('Default:', color='main'))
        for key, value in self.getDefault().items():
            if key in blacklisted_keys:
                continue
            key = key.replace('_', ' ').title()
            print(Styler.stylize(
                f'     {key}:  { "N/A" if str(value) == "" else value }', color='main'))

    def print_sorters(self):
        for i, [name, docstr, _] in enumerate(self.getSortersList()):
            print(
                Styler.stylize(
                    f'Sorter #{i + 1}, "{name}":  {f"{docstr[:300 - 3]}..." if len(docstr) > 300 else docstr}', color='main')
            )

    def print_aliases(self):
        for i, [alias_name, path] in enumerate(self.getAliasesList()):
            print(
                Styler.stylize(
                    f'Alias #{i + 1}, "{alias_name}":  {path}', color='main'
                )
            )



import os
import shutil

from appdirs import AppDirs

from helpers.styler import Styler
from helpers.exceptions import InputException, UnexpectedException
from helpers.utils import DefaultOptions, createDirsIfNotExist, getDate, print_verbose
from helpers.vars.program_constants import app_dirs

from .config.config import Config, DEFAULT_CONFIG
from .config.aliasconfig import AliasConfig
from .config.defaultconfig import DefaultConfig
from .config.sorterconfig import SorterConfig

# Check alias -> current dir -> default


class ConfigManager(Config):
    '''ConfigManager is a class that is in charge of the whole configuration. It is also the police that is in charge of making sure all of the subconfigs integrate with each other.'''
    _aliasConfig: AliasConfig
    _defaultConfig: DefaultConfig
    _sorterConfig: SorterConfig

    def __init__(self):
        config_path = os.path.join(app_dirs.user_config_dir, 'config.json')
        config_trash_dir_path = os.path.join(
            app_dirs.user_config_dir, '.trash')
        sorters_dir_path = os.path.join(app_dirs.user_config_dir, 'sorters')
        sorters_trash_dir_path = os.path.join(
            sorters_dir_path, '.trash')
        args = [config_path, app_dirs.user_config_dir, config_trash_dir_path,
                sorters_dir_path, sorters_trash_dir_path]

        print_verbose(f'Config Directory: {app_dirs.user_config_dir}')

        super(ConfigManager, self).__init__(*args)
        self._aliasConfig = AliasConfig(*args)
        self._defaultConfig = DefaultConfig(*args)
        self._sorterConfig = SorterConfig(*args)

        # Make sure all directories exist
        createDirsIfNotExist(args[1:])

        # make sure config file exist
        if not self._configExists():
            self._setFallback()

    def _trashConfig(self):
        dst_filename = f'{getDate()}.json'
        trashpath = os.path.join(
            self._config_trash_dir_path, dst_filename)
        shutil.move(self._config_path, trashpath)
        return trashpath

    def _setFallback(self):
        if self._configExists():
            sorters = self.getSortersList()
            try:
                self._trashConfig()
            except Exception as e:
                raise UnexpectedException(
                    'Unable to move config file to trash.', f'\nOriginal Error:\n       {e}')
            try:
                sorter_py_paths = [
                    sorter[2] for sorter in sorters if sorter[0] != 'basic' and sorter[0] != 'tags'
                ]
                self._sorterConfig.trashPyFiles(sorter_py_paths)
            except Exception as e:
                raise UnexpectedException(
                    'Unable to move sorters to trash.', f'\n(Original Error)\n       {e}')

        self._saveConfig(DEFAULT_CONFIG)

    def setDefault(self, default_options: DefaultOptions):
        self._defaultConfig.setDefault(
            default_options)

    def addAlias(self, alias_name: str, path: str):
        self._aliasConfig.addAlias(alias_name, path)

    def deleteAlias(self, alias_name: str):
        self._aliasConfig.deleteAlias(alias_name)

    def addSorter(self, name, filepath):
        self._sorterConfig.addSorter(name, filepath)

    def deleteSorter(self, name):
        self._sorterConfig.deleteSorter(name)
        default_sorter_name = self.getDefault('sorter')
        if (default_sorter_name == name):
            self._defaultConfig.setDefault(sorter='basic')

    def reset(self):
        self._setFallback()
        print(Styler.stylize('Successfully resetted config.', color='success'))

    def download(self, dst_path):
        if os.path.isdir(dst_path):
            dst_path = os.path.join(dst_path, 'civitdl_config')

        print(Styler.stylize(
            f'Downloading zipped config to {dst_path}.zip', color='main'))
        shutil.make_archive(dst_path, 'zip',
                            self._config_dir_path)



#!/usr/bin/env python3

import sys
import traceback
from operator import itemgetter

from helpers.styler import Styler
from helpers.exceptions import UnexpectedException
from helpers.utils import set_verbose, run_verbose, print_verbose, print_exc, DefaultOptions
from civitconfig.args.argparser import get_args
from civitconfig.data.configmanager import ConfigManager


def main():
    try:
        args = get_args()
        if args['verbose']:
            set_verbose(True)
        else:
            set_verbose(False)

        config_manager = ConfigManager()
        subcommand = args['subcommand']
        print_verbose(args)

        if subcommand == 'default':
            args.pop('subcommand')
            config_manager.setDefault(DefaultOptions(
                sorter=args['sorter'],
                max_images=args['max_images'],
                api_key=args['api_key'],
                with_prompt=args['with_prompt'],
                limit_rate=args['limit_rate'],
                retry_count=args['retry_count'],
                pause_time=args['pause_time'],
                cache_mode=args['cache_mode'],
                model_overwrite=args['model_overwrite']
            ))
            config_manager.print_defaults()
        elif subcommand == 'sorter':
            if args['add'] != None:
                add_name, add_path = args['add']
                config_manager.addSorter(add_name, add_path)
            elif args['delete'] != None:
                config_manager.deleteSorter(args['delete'])
            config_manager.print_sorters()
        elif subcommand == 'alias':
            if args['add'] != None:
                add_name, add_path = args['add']
                config_manager.addAlias(add_name, add_path)
            elif args['delete'] != None:
                config_manager.deleteAlias(args['delete'])
            config_manager.print_aliases()
        elif subcommand == 'settings':
            if args['reset'] != None:
                config_manager.reset()
            elif args['download'] != None:
                config_manager.download(args['download'])
        else:
            raise UnexpectedException(
                'Unknown subcommand not caught by argparse')

    except Exception as e:
        print('---------')
        run_verbose(traceback.print_exc)
        print_exc(e)
        print('---------')



from dataclasses import dataclass
from datetime import datetime
import json
import os
import re
import sys
from typing import Callable, Dict, Iterable, List, Literal, Union
import importlib.util
import concurrent.futures
import requests
from tqdm import tqdm

from helpers.styler import Styler
from helpers.exceptions import CustomException, InputException, UnexpectedException


@dataclass
class Id:
    type: Literal['id', 'site', 'api']
    data: List[str]
    original: str

    def __init__(self, type, data: List[str], original: str):
        if type != 'id' and type != 'site' and type != 'api':
            raise UnexpectedException(
                f'Unknown type provided for Id class: {type}')

        self.type = type
        self.data = data
        self.original = original

    def __post_init__(self):
        if self.type != 'id' and self.type != 'site' and self.type != 'api':
            raise UnexpectedException(
                f'Unknown type provided for Id class: {self.type}')
        for el in self.data:
            if not isinstance(el, str):
                raise UnexpectedException(
                    f'Wrong data type for {el} in {self.data}')

        if not isinstance(self.original, str):
            raise UnexpectedException(f'Wrong data type for {self.original}')


class SourceManager:
    def __init__(self) -> None:
        pass

    def __get_comma_list(self, string: str) -> List[str]:
        return [input_str for input_str in string.replace(
            '\n', '').split(',') if input_str.strip() != '']

    def __use_parent_dir_if_exist(self, src: str, parent: Union[str, None]) -> str:
        return os.path.normpath(os.path.join(os.path.dirname(parent), src)) if parent else src

    def parse_src(self, str_li: List[str], parent: Union[str, None] = None) -> List[Id]:
        res: List[Id] = []
        for string in str_li:
            string = string.strip()

            if string.isdigit() and abs(int(string)) == int(string):
                res.append(Id('id', [string], string))
            elif len(self.__get_comma_list(string)) > 1:
                arg_str_li = self.__get_comma_list(string)
                res.extend(self.parse_src(arg_str_li))
            elif 'civitai.com/api' in string:
                version_id_regex = r'(?<=models\/)\d+'
                version_id = re.search(version_id_regex, string)
                if version_id == None:
                    err = 'Incorrect format for the url provided' + \
                        (f' in {parent}: ' if parent else ': ') + string
                    raise InputException(err)
                version_id = version_id.group(0)
                res.append(Id('api', [version_id], string))
            elif 'civitai.com/models' in string:
                model_id = re.search(r'(?<=models\/)\d+', string)
                version_id = re.search(r'(?<=modelVersionId=)\d+', string)

                if model_id == None:
                    err = 'Incorrect format for the url/id provided' + \
                        (f' in {parent}: ' if parent else ': ') + string
                    raise InputException(err)
                model_id = model_id.group(0)

                if version_id:
                    version_id = version_id.group(0)
                    res.append(
                        Id('site', [model_id, version_id], string))
                else:
                    res.append(Id('site', [model_id], string))
            elif os.path.exists(self.__use_parent_dir_if_exist(string, parent)):
                string = self.__use_parent_dir_if_exist(string, parent)
                print(f'after: {string}')
                file_str = None

                with open(string, 'r') as file:
                    file_str = file.read().strip()
                if file_str == None:
                    raise UnexpectedException(
                        'Unknown exception while reading batchfile path.')
                str_li_res = self.__get_comma_list(file_str)
                res.extend(self.parse_src(str_li_res, parent=string))
            else:
                raise InputException(
                    f'Bad source provided: {string}', f'   Batchfile Path: {parent}' if parent else None)

        return res
